// #define TRACK_DOWN_MALLOC_PROBLEM

#include "typedef.h"
#include <assert.h>
#include <malloc.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "util.h"

extern FILE	*stdoutF;

#ifdef TRACK_DOWN_MALLOC_PROBLEM
#include <crtdbg.h>
#endif

void
error_exit(int code)
{
	fprintf(stderr,"Exit code %d\n", code);
	exit(code);
}

void
outofmemory(void)
{
	fprintf(stderr, "Out of memory\n");
	error_exit(4);
}

void *
xmalloc(size_t size)
{
#ifdef TRACK_DOWN_MALLOC_PROBLEM
	static int		initialised_crt = 0;

	if (!initialised_crt) {
		printf("Performing exhaustive checking of memory heap\n");
		initialised_crt = 1;
		_CrtSetReportMode( _CRT_WARN, _CRTDBG_MODE_FILE );
		_CrtSetReportFile( _CRT_WARN, _CRTDBG_FILE_STDOUT );
		_CrtSetReportMode( _CRT_ERROR, _CRTDBG_MODE_FILE );
		_CrtSetReportFile( _CRT_ERROR, _CRTDBG_FILE_STDOUT );
		_CrtSetReportMode( _CRT_ASSERT, _CRTDBG_MODE_FILE );
		_CrtSetReportFile( _CRT_ASSERT, _CRTDBG_FILE_STDOUT );
		// Get current flag
		int tmpFlag = _CrtSetDbgFlag( _CRTDBG_REPORT_FLAG );

		// Turn on leak-checking bit
		tmpFlag |= _CRTDBG_LEAK_CHECK_DF;
	//	tmpFlag |= _CRTDBG_CHECK_CRT_DF;
		tmpFlag |= _CRTDBG_CHECK_ALWAYS_DF;

		// Set flag to the new value
		_CrtSetDbgFlag( tmpFlag );
	}
	if (!_CrtCheckMemory()) {
		assert(0);
	}
#endif

	void *retP = malloc(size);

	if (!retP) {
		fprintf(stderr, "Can't malloc %u bytes\n", size);
		outofmemory();
	}
	return retP;
}

void
xfree(void *P)
{
	free(P);
}

u4T
unicodelth(const u2T *stringP)
{
	register	const u2T	*P  = stringP;
	register	u4T			lth = 0;

	for (;*P++; ++lth);
	return lth;
}

u2T *
unicodecpy1(u2T *string1P, const char *string2P)
{
	register		  u2T	*P1 = string1P;
	register	const char	*P2 = string2P;

	while (*P1++ = *P2++);
	return P1;
}

u2T *
unicodedup1(const char *stringP)
{
	register	const char	*P = stringP;
	register	u2T			*retP;
	size_t		lth;

	lth = strlen(stringP);
	retP = (u2T *) xmalloc((lth+1) * sizeof(u2T));
	if (!retP) {
		outofmemory();
	}
	unicodecpy1(retP, stringP);
	return(retP);
}

u2T *
unicodedup(const u2T *stringP)
{
	register u2T *retP;
	int			 lth;

	lth = unicodelth(stringP);
	retP = (u2T *) xmalloc((lth+1) * sizeof(u2T));
	unicodecpy(retP, stringP);
	return(retP);
}

u2T *
unicodecpy(u2T *string1P, const u2T *string2P)
{
	register		  u2T	*P1 = string1P;
	register	const u2T	*P2 = string2P;

	while (*P1++ = *P2++);
	return P1;
}

int 
unicodecmp(const u2T *string1P, const u2T *string2P)
{
	register	const u2T	*P1 = string1P;
	register	const u2T	*P2 = string2P;

	for (;;) {
		if (!*P1) {
			if (!*P2) {
				return(0);
			}
			return(1);
		}
		if (!*P1) {
			return(1);
		}
		if (*P1 != *P2) {
			return(1);
		}
		++P1;
		++P2;
}	}

void
dump_access(u2T	access)
{
	static const char *accessName[] = {
		/* 0x0001 */	"public ",			// ACC_PUBLIC 0x0001 Visible to everyone  Class, Method, Variable 
		/* 0x0002 */	"private ",			// ACC_PRIVATE 0x0002 Visible only to the defining class  Method, Variable 
		/* 0x0004 */	"protected ",		// ACC_PROTECTED 0x0004 Visible to subclasses  Method, Variable 
		/* 0x0008 */	"static ",			// ACC_STATIC 0x0008 Variable or method is static  Method, Variable
		/* 0x0010 */	"final ",			// ACC_FINAL 0x0010 No further subclassing, overriding, or assignment after initialization Class, Method, Variable 
		/* 0x0020 */	"synchronized ",	// ACC_SYNCHRONIZED 0x0020 Wrap use in monitor lock Method 
		/* 0x0040 */	"volatile ",		// ACC_VOLATILE 0x0040 Can't cache Variable 
		/* 0x0080 */	"transient ",		// ACC_TRANSIENT 0x0080 Not to be written or read by a persistent object manager  Variable
		/* 0x0100 */	"native ",			// ACC_NATIVE 0x0100 Implemented in a language other than Java Method 
		/* 0x0200 */	"interface ",		// ACC_INTERFACE 0x0200 Is an interface  Class 
		/* 0x0400 */	"abstract ",		// ACC_ABSTRACT 0x0400 No body provided  Class, Method 
		/* 0x0800 */	"strict ",			// ACC_STRICT 0x0800 Declared strictfp; floating-point mode is FP-strict 
		/* 0x1000 */	"synthetic ",		// ACC_SYNTHETIC
		/* 0x2000 */	"annotation ",		// ACC_ANNOTATION
		/* 0x4000 */	"enum ",			// ACC_ENUM	
		/* 0x8000 */	"0x8000 "
	};
	int	mask, i;

	for (i = 15, mask = 0x8000; mask; --i, mask >>= 1) {
		if (access & mask) {
			fputs(accessName[i], stdoutF);
	}	}
}

void
dump_classname(const u2T *classnameP)
{
	const u2T	*P;
	u2T			c;

	for (P = classnameP; (c = *P) ; ++P) {
		if (c == '/') {
			c = '.';
		}
		fputc(c, stdoutF);
	}
}

int
count_parameters(const u2T *descriptorP)
{
	u2T		c;
	int		seen      = 0;
	int		parameter = 0;
	
	for (;;) {
		switch (*descriptorP++) {
		case 0:
			return(0);
		case '[':
			continue;
		case '(':
			seen = 1;
			continue;
		case ')':
			return(parameter);
		case 'L':
			for (; (c = *descriptorP) && c != ';'; ++descriptorP);
		case 'B':
		case 'C':
		case 'D':
		case 'F':
		case 'I':
		case 'J':
		case 'S':
		case 'Z':
		case 'V':
			break;
		case ';':
			continue;
		default:
			assert(0);
		}
		if (seen) {
			++parameter;
}	}	}

void
deslash(u2T	*stringP)
{
	u2T	*P;

	for (P = stringP; ; ++P) {
		switch (*P) {
		case '/':
			*P = '.';
			break;
		case 0:
			return;
}	}	}


